#!/usr/bin/python3

#
# This script takes as input a MallocTag JSON snapshot and will perform some postprocessing.
# Available postprocessing options right now are:
#  * tree aggregation based on the thread name regex
#
# Author: fmontorsi
# Created: Oct 2023
# License: Apache license
#

import argparse
import json
import os
import sys
from mtag_node import *
from mtag_tree import *
from mtag_snapshot import *

# =======================================================================================================
# GLOBALs
# =======================================================================================================

THIS_SCRIPT_VERSION = "0.0.1"

# =======================================================================================================
# MAIN HELPERS
# =======================================================================================================


def parse_command_line():
    """Parses the command line and returns the configuration as dictionary object."""
    parser = argparse.ArgumentParser(
        description="Utility to render JSON snapshots produced by the malloc-tag library in Graphviz DOT language."
    )

    # Optional arguments
    # NOTE: we cannot add required=True to --output option otherwise it's impossible to invoke this tool with just --version
    parser.add_argument(
        "-o",
        "--output",
        help="The name of the output Graphviz DOT file with aggregated stats.",
        default=None,
    )
    parser.add_argument(
        "-v", "--verbose", help="Be verbose.", action="store_true", default=False
    )
    parser.add_argument(
        "-V",
        "--version",
        help="Print version and exit",
        action="store_true",
        default=False,
    )
    # NOTE: we use nargs='?' to make it possible to invoke this tool with just --version
    parser.add_argument(
        "input",
        nargs="?",
        help="The JSON file to analyze. If '-' the JSON is read from stdin.",
        default=None,
    )

    if "COLUMNS" not in os.environ:
        os.environ["COLUMNS"] = "120"  # avoid too many line wraps
    args = parser.parse_args()

    global verbose
    verbose = args.verbose

    if args.version:
        print(f"Version: {THIS_SCRIPT_VERSION}")
        sys.exit(0)

    if args.input is None:
        print("Please provide the input file to process as positional argument")
        parser.print_help()
        sys.exit(os.EX_USAGE)

    if args.input != "-" and not os.path.isabs(args.input):
        # take absolute path
        args.input = os.path.join(os.getcwd(), args.input)

    if args.output is not None and not os.path.isabs(args.output):
        # take absolute path
        args.output = os.path.join(os.getcwd(), args.output)

    return {
        "input_json": args.input,
        "output_file": args.output,
    }


# =======================================================================================================
# MAIN
# =======================================================================================================

if __name__ == "__main__":
    config = parse_command_line()

    # load the malloctag snapshot file:
    t = MallocTagSnapshot()
    t.load_json(config["input_json"])
    t.print_stats()

    # if requested, save the output:
    if config["output_file"]:
        t.save_graphviz_dot(config["output_file"])



#!/usr/bin/python3

# Author: fmontorsi
# Created: Oct 2023
# License: Apache license

import json
import os
import sys
import decimal
import graphviz   # pip3 install graphviz
from decimal import *

from malloctag.libs.mtag_graphviz_utils import *
from malloctag.libs.mtag_node import *

# =======================================================================================================
# GLOBALs
# =======================================================================================================

SCOPE_PREFIX = "scope_"

# =======================================================================================================
# MallocTree
# =======================================================================================================

class MallocTree:
    """
    This class represents a JSON tree produced by the malloc-tag library for an entire thread.
    """

    def __init__(self):
        self.treeRootNode = None

    def load_json(self, tree_dict: dict):
        """
        Initialize this object from the given dictionary (obtained from a malloc-tag JSON snapshot)
        """
        self.tid = tree_dict["TID"]
        self.name = tree_dict["ThreadName"]
        self.nPushNodeFailures = tree_dict["nPushNodeFailures"]
        self.nFreeTrackingFailed = tree_dict["nFreeTrackingFailed"]
        self.nMaxTreeNodes = tree_dict["nMaxTreeNodes"]
        self.nVmSizeAtCreation = tree_dict["nVmSizeAtCreation"]
        if self.nPushNodeFailures > 0:
            print(
                f"WARNING: found malloc-tag failures in tracking mem allocations inside the tree {self.name}"
            )
        if self.nFreeTrackingFailed > 0:
            print(
                f"WARNING: found malloc-tag failures in tracking mem allocations inside the tree {self.name}"
            )

        for k in tree_dict.keys():
            if k.startswith(SCOPE_PREFIX):
                # found the root node
                # print(k)
                assert self.treeRootNode is None
                self.treeRootNode = MallocTagNode(owner_tid=self.tid)
                self.treeRootNode.load_json(tree_dict[k], k[len(SCOPE_PREFIX) :])

    def get_as_dict(self):
        d = {
            "TID": self.tid,
            "ThreadName": self.name,
            "nTreeLevels": self.get_num_levels(),
            "nTreeNodesInUse": self.get_num_nodes(),
            "nMaxTreeNodes": self.nMaxTreeNodes,
            "nPushNodeFailures": self.nPushNodeFailures,
            "nFreeTrackingFailed": self.nFreeTrackingFailed,
            "nVmSizeAtCreation": self.nVmSizeAtCreation,
        }
        d[SCOPE_PREFIX + self.treeRootNode.name] = self.treeRootNode.get_as_dict()
        return d

    def save_as_graphviz_dot(self, graph):
        # let's use the digraph/subgraph label to convey extra info about this MallocTree:
        nTreeNodesInUse = self.get_num_nodes()
        labels = []
        labels.append(f"TID={self.tid}")
        labels.append(f"nPushNodeFailures={self.nPushNodeFailures}")
        labels.append(f"nTreeNodesInUse/Max={nTreeNodesInUse}/{self.nMaxTreeNodes}")

        # create one graph for each MallocTree
        tree_graph = graphviz.Digraph(name=f"cluster_TID{self.tid}")
        tree_graph.attr(label='\n'.join(labels))
        tree_graph.attr(fontsize='20')
        self.treeRootNode.save_as_graphviz_dot(tree_graph)

        # finally add the graph into the "big one" as subgraph
        graph.subgraph(tree_graph)

    def aggregate_with(self, other: "MallocTree"):
        self.tid = -1  # the TID makes no sense anymore
        self.name += "," + other.name
        self.nPushNodeFailures += other.nPushNodeFailures
        self.nFreeTrackingFailed += other.nFreeTrackingFailed
        self.nMaxTreeNodes = max(self.nMaxTreeNodes, other.nMaxTreeNodes)
        self.nVmSizeAtCreation = max(self.nVmSizeAtCreation, other.nVmSizeAtCreation)
        self.treeRootNode.aggregate_with(other.treeRootNode)

    def collect_allocated_freed_recursively(self):
        return self.treeRootNode.collect_allocated_freed_recursively()

    def compute_node_weights_recursively(self, allTreesTotalAllocatedBytes):
        self.treeRootNode.compute_node_weights_recursively(allTreesTotalAllocatedBytes)

    def get_num_levels(self):
        return self.treeRootNode.get_num_levels()

    def get_num_nodes(self):
        return self.treeRootNode.get_num_nodes()

    def get_graphviz_root_node_name(self):
        return self.treeRootNode.get_graphviz_node_name()


#!/usr/bin/python3

#
# This script takes as input a MallocTag JSON snapshot and will perform some postprocessing.
# Available postprocessing options right now are:
#  * tree aggregation based on the thread name regex
#
# Author: fmontorsi
# Created: Oct 2023
# License: Apache license
#

import argparse
import json
import os
import sys
import re
import decimal
from decimal import *
from malloctag.libs.mtag_node import *
from malloctag.libs.mtag_tree import *
from malloctag.libs.mtag_snapshot import *

# =======================================================================================================
# GLOBALs
# =======================================================================================================

THIS_SCRIPT_VERSION = "0.0.1"


# =======================================================================================================
# PostProcessConfig
# =======================================================================================================


class PostProcessAggregationRule:
    def __init__(self, ruleIdx: int):
        self.ruleIdx = ruleIdx
        self.matching_prefix = ""

    def load(self, cfg_dict):
        self.matching_prefix = cfg_dict["aggregate_trees"]["matching_prefix"]

    def logprefix(self):
        return f"Rule#{self.ruleIdx}:"

    def apply(self, snapshot: MallocTagSnapshot):
        # TODO
        regex = re.compile(self.matching_prefix)

        matching_tids = [
            tid
            for tid in snapshot.treeRegistry
            if re.match(regex, snapshot.treeRegistry[tid].name)
        ]
        print(
            f"{self.logprefix()} Found trees matching the prefix [{self.matching_prefix}] with TIDs: {matching_tids}"
        )

        if len(matching_tids) == 0:
            print(
                f"{self.logprefix()} Could not find any tree matching the prefix [{self.matching_prefix}]"
            )
        elif len(matching_tids) == 1:
            print(
                f"{self.logprefix()} Found only 1 tree matching the prefix [{self.matching_prefix}]. Nothing to aggregate."
            )
        else:
            firstTid = matching_tids[0]
            for otherTid in matching_tids[1:]:
                snapshot.aggregate_thread_trees(firstTid, otherTid)
            print(f"{self.logprefix()} Aggregation completed.")


class PostProcessConfig:
    """
    This class represents some aggregation rules provided by the user
    """

    def __init__(self):
        self.rules = []

    def load(self, cfg_json):
        wholejson = {}
        try:
            f = open(cfg_json, "r")
            text = f.read()
            f.close()
            wholejson = json.loads(text)
        except json.decoder.JSONDecodeError as err:
            print(f"Invalid configuration JSON file '{cfg_json}': {err}")
            sys.exit(1)

        nrule = 0
        for rule in wholejson:
            if not rule.startswith("rule"):
                print(
                    f"WARN: In configuration JSON file '{cfg_json}': ignoring key not starting with [rule] prefix: '{rule}'"
                )
                continue
            if len(wholejson[rule]) != 1:
                print(
                    f"In configuration JSON file '{cfg_json}': in rule '{rule}': expected exactly 1 mode"
                )
                sys.exit(1)

            mode = next(iter(wholejson[rule]))

            if mode == "aggregate_trees":
                t = PostProcessAggregationRule(nrule)
                t.load(wholejson[rule])
                self.rules.append(t)
                nrule += 1
            else:
                print(
                    f"In configuration JSON file '{cfg_json}': in rule '{rule}': found unsupported mode '{mode}'"
                )
                sys.exit(1)
        print(
            f"Loaded {len(self.rules)} postprocessing rules from config file '{cfg_json}'."
        )

    def apply(self, snapshot: MallocTagSnapshot):
        if len(self.rules) == 0:
            print(
                f"No postprocessing rules specified (see --config). The malloc-tag snapshot will not be manipulated."
            )
            return
        for r in self.rules:
            r.apply(snapshot)


# =======================================================================================================
# MAIN HELPERS
# =======================================================================================================


def parse_command_line():
    """Parses the command line and returns the configuration as dictionary object."""
    parser = argparse.ArgumentParser(
        description="Utility to post-process JSON snapshots produced by the malloc-tag library."
    )

    # Optional arguments
    # NOTE: we cannot add required=True to --output option otherwise it's impossible to invoke this tool with just --version
    parser.add_argument(
        "-o",
        "--output",
        help="The name of the output JSON file with aggregated stats.",
        default=None,
    )
    parser.add_argument(
        "-c",
        "--config",
        help="JSON file specifying the postprocessing configuration.",
        default=None,
    )
    parser.add_argument(
        "-v", "--verbose", help="Be verbose.", action="store_true", default=False
    )
    parser.add_argument(
        "-V",
        "--version",
        help="Print version and exit",
        action="store_true",
        default=False,
    )
    # NOTE: we use nargs='?' to make it possible to invoke this tool with just --version
    parser.add_argument(
        "input",
        nargs="?",
        help="The JSON file to analyze. If '-' the JSON is read from stdin.",
        default=None,
    )

    if "COLUMNS" not in os.environ:
        os.environ["COLUMNS"] = "120"  # avoid too many line wraps
    args = parser.parse_args()

    global verbose
    verbose = args.verbose

    if args.version:
        print(f"Version: {THIS_SCRIPT_VERSION}")
        sys.exit(0)

    if args.input is None:
        print("Please provide the input file to process as positional argument")
        parser.print_help()
        sys.exit(os.EX_USAGE)

    if args.input != "-" and not os.path.isabs(args.input):
        # take absolute path
        args.input = os.path.join(os.getcwd(), args.input)

    if args.output is not None and not os.path.isabs(args.output):
        # take absolute path
        args.output = os.path.join(os.getcwd(), args.output)

    return {
        "input_json": args.input,
        "output_file": args.output,
        "postprocess_config_file": args.config,
    }


# =======================================================================================================
# MAIN
# =======================================================================================================

if __name__ == "__main__":
    config = parse_command_line()

    # load the malloctag snapshot file:
    t = MallocTagSnapshot()
    t.load_json(config["input_json"])
    t.print_stats()

    # load postprocessing cfg:
    r = PostProcessConfig()
    if config["postprocess_config_file"]:
        r.load(config["postprocess_config_file"])

    # apply cfg:
    r.apply(t)

    # if requested, save the output:
    if config["output_file"]:
        t.save_json(config["output_file"])



#!/usr/bin/python3

# Author: fmontorsi
# Created: Oct 2023
# License: Apache license

import json
import os
import sys
import decimal
import graphviz   # pip3 install graphviz
from decimal import *

from malloctag.libs.mtag_graphviz_utils import *
from malloctag.libs.mtag_tree import *

# =======================================================================================================
# GLOBALs
# =======================================================================================================

TREE_PREFIX = "tree_for_TID"

# =======================================================================================================
# DecimalEncoder
# =======================================================================================================

class DecimalEncoder(json.JSONEncoder):
    def default(self, obj):
        # 👇️ if passed in object is instance of Decimal
        # convert it to a string
        if isinstance(obj, Decimal):
            return round(float(obj), 2)
        # 👇️ otherwise use the default behavior
        return json.JSONEncoder.default(self, obj)

# =======================================================================================================
# MallocTagSnapshot
# =======================================================================================================

class MallocTagSnapshot:
    """
    This class represents a JSON snapshot produced by the malloc-tag library for an entire application.
    """

    def __init__(self):
        self.treeRegistry = {}  # dict indexed by TID

    def __expand(self, snapshot_dict: dict):
        self.pid = snapshot_dict["PID"]
        self.tmStartProfiling = snapshot_dict["tmStartProfiling"]
        self.tmCurrentSnapshot = snapshot_dict["tmCurrentSnapshot"]
        self.nBytesAllocBeforeInit = snapshot_dict["nBytesAllocBeforeInit"]
        self.nBytesMallocTagSelfUsage = snapshot_dict["nBytesMallocTagSelfUsage"]
        self.vmSizeNowBytes = snapshot_dict["vmSizeNowBytes"]
        self.vmRSSNowBytes = snapshot_dict["vmRSSNowBytes"]
        self.nTotalTrackedBytes = snapshot_dict["nTotalTrackedBytes"]

        for thread_tree in snapshot_dict.keys():
            if thread_tree.startswith("tree_for_"):
                t = MallocTree()
                t.load_json(snapshot_dict[thread_tree])
                assert t.tid not in self.treeRegistry
                self.treeRegistry[t.tid] = t

    def load_json(self, json_infile: str):
        # load the JSON
        wholejson = {}
        try:
            f = open(json_infile, "r")
            text = f.read()
            f.close()
            wholejson = json.loads(text)
        except json.decoder.JSONDecodeError as err:
            print("Invalid input JSON file '%s': %s" % (json_infile, err))
            sys.exit(1)

        # process it
        self.__expand(wholejson)

        # recompute weights
        self.compute_node_weights_recursively()

    def save_json(self, json_outfile: str):
        outdict = {
            "PID": self.pid,
            "tmStartProfiling": self.tmStartProfiling,
            "tmCurrentSnapshot": self.tmCurrentSnapshot,
        }
        for t in self.treeRegistry.keys():
            outdict[TREE_PREFIX + str(t)] = self.treeRegistry[t].get_as_dict()

        # add last few properties:
        outdict["nBytesAllocBeforeInit"] = self.nBytesAllocBeforeInit
        outdict["nBytesMallocTagSelfUsage"] = self.nBytesMallocTagSelfUsage
        outdict["vmSizeNowBytes"] = self.vmSizeNowBytes
        outdict["vmRSSNowBytes"] = self.vmRSSNowBytes
        outdict["nTotalTrackedBytes"] = self.nTotalTrackedBytes

        getcontext().prec = 2
        with open(json_outfile, "w", encoding="utf-8") as f:
            json.dump(outdict, f, ensure_ascii=False, indent=4, cls=DecimalEncoder)
        print(f"Saved postprocessed results into {json_outfile}.")

    def save_graphviz_dot(self, graphviz_dot_outfile: str):
        thegraph = graphviz.Digraph(comment='Malloc-tag snapshot')

        labels = []
        labels.append("Whole process stats")
        labels.append(
            "allocated_mem_before_malloctag_init=" + GraphVizUtils.pretty_print_bytes(self.nBytesAllocBeforeInit))
        labels.append(
            "allocated_mem_by_malloctag_itself=" + GraphVizUtils.pretty_print_bytes(self.nBytesMallocTagSelfUsage))
        labels.append("allocated_mem=" + GraphVizUtils.pretty_print_bytes(self.nTotalTrackedBytes))
        labels.append("vm_size_now=" + GraphVizUtils.pretty_print_bytes(self.vmSizeNowBytes))
        labels.append("vm_rss_now=" + GraphVizUtils.pretty_print_bytes(self.vmRSSNowBytes))
        labels.append("malloctag_start_ts=" + self.tmStartProfiling)
        labels.append("this_snapshot_ts=" + self.tmCurrentSnapshot)

        # create the main node:
        mainNodeName = f"Process_{self.pid}"
        thegraph.node(mainNodeName, label='\n'.join(labels))

        # used to compute weights later:
        totalloc, totfreed = self.collect_allocated_freed_recursively()

        # now create subgraphs for each and every tree: 
        for t in self.treeRegistry.keys():
            self.treeRegistry[t].save_as_graphviz_dot(thegraph)

            # compute the weight for the 't'-th tree:
            treealloc, treefree = self.treeRegistry[t].collect_allocated_freed_recursively()
            w = 0 if totalloc == 0 else 100*treealloc/totalloc
            wstr = f"%.2f%%" % w

            # add edge main node ---> tree
            thegraph.edge(mainNodeName, self.treeRegistry[t].get_graphviz_root_node_name(), label=wstr)


        #print(thegraph.source)
        thegraph.render(outfile=graphviz_dot_outfile)

    def print_stats(self):
        num_nodes = sum(
            [self.treeRegistry[t].get_num_nodes() for t in self.treeRegistry]
        )
        print(
            f"Loaded a total of {len(self.treeRegistry)} trees containing {num_nodes} nodes."
        )

    def aggregate_thread_trees(self, tid1: int, tid2: int):
        # do the aggregation
        self.treeRegistry[tid1].aggregate_with(self.treeRegistry[tid2])
        # remove the aggregated tree:
        del self.treeRegistry[tid2]
        # update all node weights:
        self.compute_node_weights_recursively()


    def collect_allocated_freed_recursively(self):
        totalloc = 0
        totfreed = 0
        for t in self.treeRegistry.keys():
            a, f = self.treeRegistry[t].collect_allocated_freed_recursively()
            totalloc += a
            totfreed += f
        return totalloc, totfreed

    def compute_node_weights_recursively(self):
        totalloc, totfreed = self.collect_allocated_freed_recursively()

        # in each tree, recompute node weights using the total allocated as denominator:
        for t in self.treeRegistry.keys():
            self.treeRegistry[t].compute_node_weights_recursively(totalloc)



#!/usr/bin/python3

# Author: fmontorsi
# Created: Oct 2023
# License: Apache license

# =======================================================================================================
# GraphVizUtils
# =======================================================================================================

class GraphVizUtils:
    """
    This class provides utilities for producing .DOT graphviz outputs
    """

    def __init__(self):
        pass

    @staticmethod
    def pretty_print_bytes(bytes):
        # NOTE: we convert to kilo/mega/giga (multiplier=1000) not to kibi/mebi/gibi (multiplier=1024) bytes !!!
        if bytes < 1000:
            return str(bytes) + "B"
        elif bytes < 1000000:
            return str(bytes // 1000) + "kB"
        elif bytes < 1000000000:
            return str(bytes // 1000000) + "MB"
        else:
            return str(bytes // 1000000000) + "GB"




#!/usr/bin/python3

# Author: fmontorsi
# Created: Oct 2023
# License: Apache license

import json
import os
import sys
import decimal
import graphviz   # pip3 install graphviz
from decimal import *
from malloctag.libs.mtag_graphviz_utils import *

# =======================================================================================================
# GLOBALs
# =======================================================================================================

SCOPE_PREFIX = "scope_"
g_num_nodes = 0


# =======================================================================================================
# MallocTagNode
# =======================================================================================================

class MallocTagNode:
    """
    This class represents a node in the JSON tree produced by the malloc-tag library.
    """

    def __init__(self, owner_tid, level=1):
        global g_num_nodes
        self.childrenNodes = {}  # dict indexed by "scope name"

        # allocate a new ID for this node
        self.id = g_num_nodes
        g_num_nodes += 1

        # a few info coming from the "caller"
        self.nLevel = level
        self.ownerTID = owner_tid

    def load_json(self, node_dict: dict, name: str):
        # this is the "scope name"
        self.name = name

        # weights
        self.nTotalWeightPercentage = 0 # this will be recomputed on the fly later
        self.nSelfWeightPercentage = 0 # this will be recomputed on the fly later

        # load self properties
        self.nBytesTotalAllocated = node_dict["nBytesTotalAllocated"]
        self.nBytesSelfAllocated = node_dict["nBytesSelfAllocated"]
        self.nBytesSelfFreed = node_dict["nBytesSelfFreed"]
        self.nTimesEnteredAndExited = node_dict["nTimesEnteredAndExited"]
        self.nCallsTo_malloc = node_dict["nCallsTo_malloc"]
        self.nCallsTo_realloc = node_dict["nCallsTo_realloc"]
        self.nCallsTo_calloc = node_dict["nCallsTo_calloc"]
        self.nCallsTo_free = node_dict["nCallsTo_free"]

        # recursive load
        for scope in node_dict["nestedScopes"].keys():
            assert scope.startswith(SCOPE_PREFIX)
            name = scope[len(SCOPE_PREFIX) :]

            # load the node recursively
            t = MallocTagNode(owner_tid=self.ownerTID, level=self.nLevel+1)
            t.load_json(node_dict["nestedScopes"][scope], name)
            assert name not in self.childrenNodes

            # store the node
            self.childrenNodes[name] = t

    def get_as_dict(self):
        d = {
            "nBytesTotalAllocated": self.nBytesTotalAllocated,
            "nBytesSelfAllocated": self.nBytesSelfAllocated,
            "nBytesSelfFreed": self.nBytesSelfFreed,
            "nTimesEnteredAndExited": self.nTimesEnteredAndExited,
            # FIXME: rename to nTotalWeightPercentage also in the JSON output
            "nWeightPercentage": Decimal(self.nTotalWeightPercentage), # see usage of DecimalEncoder later on
            "nCallsTo_malloc": self.nCallsTo_malloc,
            "nCallsTo_realloc": self.nCallsTo_realloc,
            "nCallsTo_calloc": self.nCallsTo_calloc,
            "nCallsTo_free": self.nCallsTo_free,
            "nestedScopes": {},
        }
        for n in self.childrenNodes:
            d["nestedScopes"][SCOPE_PREFIX + n] = self.childrenNodes[n].get_as_dict()
        return d

    def save_as_graphviz_dot(self, graph):

        thisNodeShape = "ellipse"
        thisNodeLabels = []
        if self.nLevel == 1:
            thisNodeLabels.append(f"thread={self.name}")
            thisNodeLabels.append(f"TID={self.ownerTID}")
            thisNodeShape = "box"
        else:
            thisNodeLabels.append(f"scope={self.name}")

        # for each node provide an overall view of
        # - total memory usage accounted for this node (both in bytes and as percentage)
        # - self memory usage (both in bytes and as percentage)
        if self.nBytesSelfAllocated != self.nBytesTotalAllocated:
            thisNodeLabels.append(f"total_alloc={GraphVizUtils.pretty_print_bytes(self.nBytesTotalAllocated)}")
            thisNodeLabels.append(f"self_alloc={GraphVizUtils.pretty_print_bytes(self.nBytesSelfAllocated)}")
        else:
            # shorten the label:
            thisNodeLabels.append(f"total_alloc=self_alloc={GraphVizUtils.pretty_print_bytes(self.nBytesTotalAllocated)}")
        
        thisNodeLabels.append(f"self_freed={GraphVizUtils.pretty_print_bytes(self.nBytesSelfFreed)}")
        thisNodeLabels.append(f"visited_times={self.nTimesEnteredAndExited}")

        n = self.get_avg_self_bytes_alloc_per_visit()
        thisNodeLabels.append(f"self_alloc_per_visit={GraphVizUtils.pretty_print_bytes(n)}")

        thisNodeLabels.append(f"nCallsTo_malloc={self.nCallsTo_malloc}")
        thisNodeLabels.append(f"nCallsTo_realloc={self.nCallsTo_realloc}")
        thisNodeLabels.append(f"nCallsTo_calloc={self.nCallsTo_calloc}")
        thisNodeLabels.append(f"nCallsTo_free={self.nCallsTo_free}")

        # Calculate the fillcolor in a range from 0-9 based on the "self weight"
        # The idea is to provide an intuitive indication of the self contributions of each malloc scope:
        # The bigger/eye-catching nodes will be those where a lot of byte allocations have been recorded,
        # regardless of what happened inside their children
        thisNodeFillColor = ""
        thisNodeFontSize = "14"

        if self.nSelfWeightPercentage < 5:
            thisNodeFillColor = "1"
            thisNodeFontSize = "9"
        elif self.nSelfWeightPercentage < 10:
            thisNodeFillColor = "2"
            thisNodeFontSize = "10"
        elif self.nSelfWeightPercentage < 20:
            thisNodeFillColor = "3"
            thisNodeFontSize = "12"
        elif self.nSelfWeightPercentage < 40:
            thisNodeFillColor = "4"
            thisNodeFontSize = "14"
        elif self.nSelfWeightPercentage < 60:
            thisNodeFillColor = "5"
            thisNodeFontSize = "16"
        elif self.nSelfWeightPercentage < 80:
            thisNodeFillColor = "6"
            thisNodeFontSize = "18"
        else:
            thisNodeFillColor = "7"
            thisNodeFontSize = "20"
        
        # create a name that is unique in the whole graphviz DOT document:
        thisNodeName = self.get_graphviz_node_name()

        # finally add this node:
        graph.node(name=thisNodeName, label='\n'.join(thisNodeLabels), shape=thisNodeShape, fillcolor=thisNodeFillColor, fontsize=thisNodeFontSize)
            
        # write all the connections between this node and its children:
        for c in self.childrenNodes:
            edge_label = f"w={self.childrenNodes[c].nTotalWeightPercentage}%"
            graph.edge(thisNodeName, self.childrenNodes[c].get_graphviz_node_name(), label=edge_label);

        # now recurse into each children:
        for c in self.childrenNodes:
            self.childrenNodes[c].save_as_graphviz_dot(graph)

    def aggregate_with(self, other: "MallocTagNode"):
        # reset weights... they will need to be recomputed:
        self.nTotalWeightPercentage = 0
        self.nSelfWeightPercentage = 0

        # sum all "summable" properties
        self.nBytesTotalAllocated += other.nBytesTotalAllocated
        self.nBytesSelfAllocated += other.nBytesSelfAllocated
        self.nBytesSelfFreed += other.nBytesSelfFreed
        self.nTimesEnteredAndExited += other.nTimesEnteredAndExited
        self.nCallsTo_malloc += other.nCallsTo_malloc
        self.nCallsTo_realloc += other.nCallsTo_realloc
        self.nCallsTo_calloc += other.nCallsTo_calloc
        self.nCallsTo_free += other.nCallsTo_free

        # recurse into children:
        for scopeName in other.childrenNodes.keys():
            if scopeName in self.childrenNodes:
                # same scope is already present... aggregate!
                self.childrenNodes[scopeName].aggregate_with(
                    other.childrenNodes[scopeName]
                )
            else:
                # this is a new scope... present only in the 'other' node... add it
                self.childrenNodes[scopeName] = other.childrenNodes[scopeName]



    def get_num_levels(self):
        tot = 1
        if self.childrenNodes:
            # take the max from children nodes:
            tot += max(
                [self.childrenNodes[n].get_num_levels() for n in self.childrenNodes]
            )
        return tot

    def get_num_nodes(self):
        tot = 1  # this current node
        # recurse into children nodes:
        for t in self.childrenNodes.keys():
            tot += self.childrenNodes[t].get_num_nodes()
        return tot

    def get_avg_self_bytes_alloc_per_visit(self):
        if self.nTimesEnteredAndExited > 0:
            # it's questionable if we should instead use:
            #                get_net_self_bytes() / m_nTimesEnteredAndExited
            return self.nBytesSelfAllocated / self.nTimesEnteredAndExited
        return 0

    def get_graphviz_node_name(self):
        # create a name that is unique in the whole graphviz DOT document:
        return f"{self.ownerTID}_{self.name}"


    def collect_allocated_freed_recursively(self):
        # Postorder traversal of a tree:
        # First, traverse all children subtrees:
        alloc_bytes = 0
        freed_bytes = 0
        for child in self.childrenNodes:
            a, f = self.childrenNodes[child].collect_allocated_freed_recursively()
            alloc_bytes += a
            freed_bytes += f

        # Finally, "visit" this node, updating the bytes count, using all children contributions:
        self.nBytesTotalAllocated = alloc_bytes + self.nBytesSelfAllocated
        self.nBytesTotalFreed = freed_bytes + self.nBytesSelfFreed

        # Assume the provided pointers had been initialized to zero at the start of the recursion
        return self.nBytesTotalAllocated, self.nBytesTotalFreed

    def compute_node_weights_recursively(self, allTreesTotalAllocatedBytes):
        if allTreesTotalAllocatedBytes == 0:
            self.nTotalWeightPercentage = self.nSelfWeightPercentage = 0
        else:
            # Compute weight of this node:
            self.nTotalWeightPercentage = round(float(100 * self.nBytesTotalAllocated) / allTreesTotalAllocatedBytes, 2)
            self.nWeightSelf = round(float(100 * self.nBytesSelfAllocated) / allTreesTotalAllocatedBytes, 2)

        # recurse:
        for child in self.childrenNodes:
            self.childrenNodes[child].compute_node_weights_recursively(allTreesTotalAllocatedBytes)







